<?php

namespace Drupal\access_policy;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\AlterableInterface;
use Drupal\Core\Database\Query\ConditionInterface;
use Drupal\Core\Database\Query\SelectInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Session\AccountProxy;
use Drupal\access_policy\Entity\AccessPolicyInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class for handling Access Policy query alter.
 *
 * This generates a dynamic query based on how access policies and access
 * rules are configured. The logic of the select query is always going to follow
 * this format.
 *
 * SELECT column FROM node_field_data
 * WHERE
 * access_policy = 'policy' AND (field = 0 AND/OR field2 = 3...) OR
 * access_policy = 'policy 2' AND (field = 1 AND/OR field2 = 3...)
 *
 * @package Drupal\access_policy
 */
class AccessPolicyQueryAlter implements ContainerInjectionInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The access rule plugin manager.
   *
   * @var \Drupal\access_policy\AccessRulePluginManager
   */
  protected $accessRuleManager;

  /**
   * The current_user service.
   *
   * @var \Drupal\Core\Session\AccountProxy
   */
  protected $currentUser;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The base table name.
   *
   * @var string
   */
  protected $baseTable;

  /**
   * The base field name.
   *
   * @var string
   */
  protected $baseField;

  /**
   * AccessPolicyQueryAlter constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\access_policy\AccessRulePluginManager $access_rule_manager
   *   The access rule plugin manager.
   * @param \Drupal\Core\Session\AccountProxy $current_user
   *   The current_user service.
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, AccessRulePluginManager $access_rule_manager, AccountProxy $current_user, Connection $connection) {
    $this->entityTypeManager = $entity_type_manager;
    $this->accessRuleManager = $access_rule_manager;
    $this->currentUser = $this->entityTypeManager->getStorage('user')->load($current_user->id());
    $this->connection = $connection;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.access_rule'),
      $container->get('current_user'),
      $container->get('database')
    );
  }

  /**
   * Alter a query.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   *
   * @see hook_query_alter()
   */
  public function queryAlter(AlterableInterface $query, EntityTypeInterface $entity_type) {
    $altered = $query->getMetaData('access_policy');
    // If the base table does not match the entity type then do not proceed with
    // query. This can happen in the case of books or forums which both tag the
    // query with node_access.
    // @todo add tests for books/forums.
    if (!$altered) {
      $base_table = $this->getBaseTable($query, $entity_type);

      // The table is valid, we can safely get the base field.
      $this->baseField = $entity_type->getKey('id');

      // Only query the access policies of the current active entity type.
      $access_policies = $this->entityTypeManager->getStorage('access_policy')
        ->loadByProperties(['target_entity_type_id' => $entity_type->id()]);

      $this->ensureTable($query, $entity_type);

      $or_group = $query->orConditionGroup();
      foreach ($access_policies as $access_policy) {
        // @todo this currently doesn't fully support AccessPolicyType.
        // For example, it can't accommodate "rules only" policies.
        /** @var \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy */
        $this->accessPolicyQueryAlter($access_policy, $query, $or_group);
      }

      // If any policies exist then we need to add the OR group for
      // unassigned and empty policies. If no policies exist then we
      // shouldn't modify the query at all.
      if (!empty($access_policies)) {
        $column = $this->getAccessPolicyColumn($query);
        $or_group->condition($column, NULL, 'IS NULL');
        $or_group->condition($column, '');
        $query->condition($or_group);
      }

      // Flag this query as altered by access policy.
      $query->addMetaData('access_policy', TRUE);
    }
  }

  /**
   * Get the current base table.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   *
   * @return string
   *   The current base table.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getBaseTable(AlterableInterface $query, EntityTypeInterface $entity_type) {
    $tables = $query->getTables();
    // The book and forum module both provide a base_table. If no base table
    // metadata exists then we can safely assume it's a normal entity access
    // tag query alter.
    $base_table = $query->getMetaData('base_table');
    if (!$base_table) {
      /** @var \Drupal\Core\Entity\Sql\DefaultTableMapping $table_mapping */
      $table_mapping = \Drupal::entityTypeManager()->getStorage($entity_type->id())->getTableMapping();
      $entity_base_tables = $table_mapping->getTableNames();

      foreach ($tables as $table_info) {
        if (!($table_info instanceof SelectInterface)) {
          $table = $table_info['table'];
          // If the table is either the base table or data table then use
          // one of those.
          if ($table == $entity_type->getBaseTable() || $table == $entity_type->getDataTable()) {
            // If the table has an alias then make sure and use that instead.
            $base_table = $table_info['alias'] ?? $table;
            break;
          }
          // If one of the node base tables are in the query, add it to the list
          // of possible base tables to join against.
          if (in_array($table, $entity_base_tables) && is_null($table_info['join type'])) {
            $base_table = $table_info['alias'] ?? $table;
            break;
          }
        }
      }
    }

    $this->baseTable = $base_table;
    return $this->baseTable;
  }

  /**
   * Make sure that the access policy table has been joined properly.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type.
   */
  public function ensureTable(AlterableInterface $query, EntityTypeInterface $entity_type) {
    $tables = $query->getTables();
    $join = TRUE;
    foreach ($tables as $table) {
      // Ensure that we don't join the table twice. this can happen when
      // filtering on views.
      if ($table['table'] == 'content_access_policy') {
        $join = FALSE;
        break;
      }
    }

    if ($join) {
      // Always join the access policy table.
      $join_condition = "content_access_policy.content_entity_id = " . $this->getBaseTablePlaceholder() . " AND content_access_policy.content_entity_type_id = '" . $entity_type->id() . "'";
      $query->leftJoin('content_access_policy', 'content_access_policy', $join_condition);
    }
  }

  /**
   * Alter a query based on Access Policies.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The Access Policy.
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query to alter.
   * @param \Drupal\Core\Database\Query\ConditionInterface $orGroup
   *   The condition group.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function accessPolicyQueryAlter(AccessPolicyInterface $access_policy, AlterableInterface $query, ConditionInterface $orGroup) {
    $access_policy_id = $access_policy->id();

    if ($this->userCanView($access_policy, $this->currentUser)) {
      // The access policy will always have the logic of
      // OR access_policy = 'policy' AND (access rules AND/OR)
      $has_access_rules = $this->hasAccessRules($access_policy);
      if ($has_access_rules) {
        $operator = $access_policy->getAccessRuleOperator();
        switch ($operator) {
          case 'AND':
            $this->createAndGroupCondition($access_policy, $query, $orGroup);
            break;

          case 'OR':
            $this->createOrGroupCondition($access_policy, $query, $orGroup);
            break;
        }
      }

      // If no rules are supported or no query was altered then do a normal
      // condition without any AND group.
      if (!$has_access_rules || !$this->queryIsAltered($orGroup)) {
        $orGroup->condition($this->getAccessPolicyColumn($query), $access_policy_id);
      }
    }
  }

  /**
   * Get the base table placeholder.
   *
   * @return string
   *   The base table.
   */
  private function getBaseTablePlaceholder() {
    return $this->baseTable . '.' . $this->baseField;
  }

  /**
   * Determine whether the access policy has rules that should be queried.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   *
   * @return bool
   *   TRUE if it has access rules; FALSE otherwise.
   */
  private function hasAccessRules(AccessPolicyInterface $access_policy) {
    $support_rules = $access_policy->getTypePlugin()->supportsRules();
    $has_rules = $access_policy->getAccessRules();
    $does_not_bypass = !$this->currentUser->hasPermission("bypass {$access_policy->id()} access rules");
    if ($support_rules && $has_rules && $does_not_bypass) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Create the And group condition for access rule plugins.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The Access Policy.
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query to alter.
   * @param \Drupal\Core\Database\Query\ConditionInterface $orGroup
   *   The condition group.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function createAndGroupCondition(AccessPolicyInterface $access_policy, AlterableInterface $query, ConditionInterface $orGroup) {
    $access_rules = $this->getAccessRulePlugins($access_policy);
    $andGroup = $query->andConditionGroup();

    $this->accessRulesQueryAlter($access_rules, $query, $andGroup);

    // If the query was actually altered then create the AND group and add the
    // conditions. This helps to prevent plugins without queries from breaking
    // the query.
    if ($this->queryIsAltered($andGroup)) {
      $andGroup->condition($this->getAccessPolicyColumn($query), $access_policy->id());

      $orGroup->condition($andGroup);
    }
  }

  /**
   * Create the OR group condition for access rule plugins.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The Access Policy.
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query to alter.
   * @param \Drupal\Core\Database\Query\ConditionInterface $orGroup
   *   The condition group.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  private function createOrGroupCondition(AccessPolicyInterface $access_policy, AlterableInterface $query, ConditionInterface $orGroup) {
    $access_rules = $this->getAccessRulePlugins($access_policy);
    $condGroup = $query->orConditionGroup();

    $this->accessRulesQueryAlter($access_rules, $query, $condGroup);

    // If the query was actually altered then create the AND group and add the
    // OR conditions. This helps to prevent plugins without queries from
    // breaking the query.
    if ($this->queryIsAltered($condGroup)) {
      $andGroup = $query->andConditionGroup();
      $andGroup->condition($condGroup);
      $andGroup->condition($this->getAccessPolicyColumn($query), $access_policy->id());

      $orGroup->condition($andGroup);
    }
  }

  /**
   * Get the access rule plugins.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   *
   * @return array
   *   Array of access rule plugins.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getAccessRulePlugins(AccessPolicyInterface $access_policy) {
    $plugins = $this->accessRuleManager->getAccessRulePlugins($access_policy);

    return array_filter($plugins, function ($plugin) {
      $config = $plugin->getConfiguration();
      if ($config['query']) {
        return TRUE;
      }
      return FALSE;
    });
  }

  /**
   * Determine whether the query was actually altered.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $condGroup
   *   The condition group.
   *
   * @return bool
   *   TRUE if the query was altered, FALSE otherwise.
   */
  private function queryIsAltered(ConditionInterface $condGroup) {
    foreach ($condGroup->conditions() as $key => $condition) {
      if (is_int($key)) {
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Alter a query based on Access Rules.
   *
   * @param array $access_rules
   *   The Access Rules.
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query to alter.
   * @param \Drupal\Core\Database\Query\ConditionInterface $condGroup
   *   The condition group.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   */
  public function accessRulesQueryAlter(array $access_rules, AlterableInterface $query, ConditionInterface $condGroup) {
    foreach ($access_rules as $access_rule) {
      $select = new SelectQueryStub($this->connection, $this->baseTable);
      $select->addMetaData('base_table', $this->baseTable);
      $select->addMetaData('base_field', $this->baseField);

      $access_rule->queryAlter($select, $this->currentUser);

      if ($select->isDistinct()) {
        $query->distinct();
      }

      $this->joinTablesFromQueryStub($query, $select);
      $this->addConditionsFromQueryStub($condGroup, $select);
    }
  }

  /**
   * Join tables from the query stub to the main query.
   *
   * This also ensures that any tables that have already been joined don't get
   * joined again as it could create duplicate results.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The main query.
   * @param \Drupal\Core\Database\Query\SelectInterface $queryStub
   *   The select query stub.
   */
  public function joinTablesFromQueryStub(AlterableInterface $query, SelectInterface $queryStub) {
    $existing_tables = $query->getTables();
    $tables = $queryStub->getTables();
    $join_tables = array_filter($tables, function ($new_table) use ($existing_tables) {
      if (!is_null($new_table['join type'])) {
        $join_table = TRUE;
        // Make sure we don't join tables that have already been joined.
        foreach ($existing_tables as $table) {
          if ($new_table['table'] == $table['table'] && $new_table['alias'] == $table['alias']) {
            $join_table = FALSE;
            break;
          }
        }
        return $join_table;
      }
      return FALSE;
    });

    foreach ($join_tables as $table) {
      $query->addJoin($table['join type'], $table['table'], $table['alias'], $table['condition'], $table['arguments']);
    }
  }

  /**
   * Add conditions from a pseudo query.
   *
   * @param \Drupal\Core\Database\Query\ConditionInterface $condGroup
   *   The condition group.
   * @param \Drupal\Core\Database\Query\SelectInterface $queryStub
   *   The select query stub.
   */
  public function addConditionsFromQueryStub(ConditionInterface $condGroup, SelectInterface $queryStub) {
    $conditions = $queryStub->conditions();

    foreach ($conditions as $key => $condition) {
      if (is_int($key)) {
        $condGroup->condition($condition['field'], $condition['value'], $condition['operator']);
      }
    }
  }

  /**
   * Get Access Strategy Column Name based on the table name.
   *
   * @param \Drupal\Core\Database\Query\AlterableInterface $query
   *   The query.
   *
   * @return string
   *   The column name.
   */
  protected function getAccessPolicyColumn(AlterableInterface $query): string {
    return 'content_access_policy.access_policy';
  }

  /**
   * Determine whether the current user can view content with an access policy.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $access_policy
   *   The access policy.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   *
   * @return bool
   *   TRUE if the user has access; FALSE otherwise.
   */
  private function userCanView(AccessPolicyInterface $access_policy, AccountInterface $account) {
    if ($access_policy->getTypePlugin()->shouldValidatePermission('view')) {
      $entity_type = $access_policy->getTargetEntityType();
      $label = strtolower($entity_type->getLabel());
      return $account->hasPermission("view {$access_policy->id()} " . $label);
    }

    return TRUE;
  }

}
