<?php

namespace Drupal\access_policy;

use Drupal\access_policy\Entity\AccessPolicyInterface;
use Drupal\access_policy\Plugin\AccessPolicyType\AccessPolicyTypeInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Session\AccountInterface;

/**
 * Provides an Access Policy validator service.
 */
class AccessRuleValidator {

  /**
   * The ContentAccessPolicyManager service.
   *
   * @var \Drupal\access_policy\ContentAccessPolicyManager
   */
  protected $contentAccessPolicyManager;

  /**
   * The AccessRulePluginManager service.
   *
   * @var \Drupal\access_policy\AccessRulePluginManager
   */
  protected $accessRulePluginManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a AccessPolicyEntityAccessControlHandler object.
   *
   * @param \Drupal\access_policy\ContentAccessPolicyManager $content_access_policy_manager
   *   The ContentAccessPolicyManager service.
   * @param \Drupal\access_policy\AccessRulePluginManager $access_rule_plugin_manager
   *   The AccessRulePluginManager service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager service.
   */
  public function __construct(ContentAccessPolicyManager $content_access_policy_manager, AccessRulePluginManager $access_rule_plugin_manager, EntityTypeManagerInterface $entity_type_manager) {
    $this->contentAccessPolicyManager = $content_access_policy_manager;
    $this->accessRulePluginManager = $access_rule_plugin_manager;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Validate Access Rules in a policy.
   *
   * @param \Drupal\access_policy\Entity\AccessPolicyInterface $policy
   *   The access policy.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The content entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   * @param string $operation
   *   The current operation.
   *
   * @return bool
   *   Return TRUE if valid, FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function validateAccessPolicyRules(AccessPolicyInterface $policy, EntityInterface $entity, AccountInterface $account, $operation) {
    $access_rules = $this->accessRulePluginManager->getApplicableAccessRulePlugins($policy, $entity);
    $type = $policy->getTypePlugin();
    if ($type->shouldValidateTypeRules($operation)) {
      $access_rules = $this->filterAgainstTypeRules($access_rules, $type, $operation);
    }

    return $this->validateAccessRules($access_rules, $entity, $account, $policy->getAccessRuleOperator());
  }

  /**
   * Validate Access Rules.
   *
   * @param array $access_rules
   *   Array of access rule plugins.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The content entity.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   * @param string $operator
   *   The current operator; default is AND.
   *
   * @return bool
   *   Return TRUE if valid, FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function validateAccessRules(array $access_rules, EntityInterface $entity, AccountInterface $account, $operator = 'AND') {
    $user = $this->entityTypeManager->getStorage('user')->load($account->id());
    $access = TRUE;

    foreach ($access_rules as $plugin) {
      $access = $plugin->validate($entity, $user);

      switch ($operator) {
        // If all rules are required then fail immediately if any return false.
        case 'AND':
          if (!$access) {
            return FALSE;
          }
          break;

        // If any rules are required then pass immediately if any return true.
        case 'OR':
          if ($access) {
            return TRUE;
          }
          break;
      }

    }

    return $access;
  }

  /**
   * Filter against the default rules.
   *
   * If we should be validating against a type rule instead of the ones
   * provided then this filters out all the non-type rules.
   *
   * @param array $rules
   *   Array of access rule plugins.
   * @param \Drupal\access_policy\Plugin\AccessPolicyType\AccessPolicyTypeInterface $type
   *   The access policy type interface.
   * @param string $operation
   *   The operation.
   *
   * @return array
   *   Array of access rule plugins.
   */
  private function filterAgainstTypeRules(array $rules, AccessPolicyTypeInterface $type, $operation) {
    $allowed_rules = $type->getTypeRuleIds($operation);
    $filtered_rules = [];
    foreach ($rules as $id => $plugin) {
      if (in_array($id, $allowed_rules)) {
        $filtered_rules[$id] = $plugin;
      }
    }
    return $filtered_rules;
  }

}
