<?php

namespace Drupal\Tests\access_policy\Functional;

use Drupal\access_policy\Entity\AccessPolicy;
use Drupal\user\Entity\Role;

/**
 * Ensures that listings only display entities that a user is allowed to view.
 *
 * @group access_policy
 */
class AccessPolicyEntityListingTest extends AccessPolicyTestBase {

  /**
   * Modules to enable.
   *
   * @var array
   */
  protected static $modules = [
    'access_policy',
    'access_policy_test',
    'filter',
    'node',
    'field',
    'user',
    'datetime',
    'views',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Set to TRUE to strict check all configuration saved.
   *
   * @var bool
   *
   * @see \Drupal\Core\Config\Development\ConfigSchemaChecker
   */
  protected $strictConfigSchema = FALSE;

  /**
   * Tests that a listing only displays entities that a user is allowed to view.
   */
  public function testContentListing() {
    AccessPolicy::create([
      'id' => 'matching_text',
      'label' => 'Matching Text Access Policy',
      'access_rules' =>  [
        'matching_text' => [
          'id' => 'matching_text',
          'plugin_id' => 'matching_text',
          'settings' => [
            'query' => TRUE,
          ],
        ],
      ],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ])->save();

    AccessPolicy::create([
      'id' => 'test_policy',
      'label' => 'Test Access Policy',
      'access_rules' =>  [
        'access_rule_test' => [
          'id' => 'access_rule_test',
          'plugin_id' => 'access_rule_test',
          'settings' => [
            'query' => TRUE,
          ],
        ],
      ],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ])->save();

    $admin_user = $this->drupalCreateUser([
      'access administration pages',
      'access content overview',
      'administer nodes',
      'bypass node access',
      'view matching_text content',
    ]);
    $admin_user->set('field_text', 'Matching text');
    $admin_user->save();

    $node_1 = $this->drupalCreateNode([
      'title' => 'First',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Matching text',
    ]);

    $node_2 = $this->drupalCreateNode([
      'title' => 'Second',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Restricted',
    ]);

    $node_3 = $this->drupalCreateNode([
      'title' => 'Third',
      'type' => 'page',
      'status' => 1,
    ]);

    $node_4 = $this->drupalCreateNode([
      'title' => 'Fourth',
      'type' => 'page',
      'status' => 1,
    ]);

    $this->assignAccessPolicy($node_1, 'matching_text');
    $this->assignAccessPolicy($node_2, 'matching_text');
    $this->assignAccessPolicy($node_3, 'matching_text');
    $this->assignAccessPolicy($node_4, 'test_policy');

    $this->drupalLogin($admin_user);

    // Verify that only the content that the user can access is listed.
    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('First');
    $this->assertSession()->pageTextNotContains('Second');
    $this->assertSession()->pageTextNotContains('Third');
    $this->assertSession()->pageTextNotContains('Fourth');
  }

  /**
   * Tests that a listing only displays entities that a user is allowed to view
   * when that policy has multiple rules using AND.
   */
  public function testContentListingWithMultipleANDRules() {
    AccessPolicy::create([
      'id' => 'test_policy_with_and_rules',
      'label' => 'Test Access Policy with AND rules',
      'access_rules' =>  [
        'matching_text' => [
          'id' => 'matching_text',
          'plugin_id' => 'matching_text',
          'settings' => [
            'query' => TRUE,
          ],
        ],
        'is_own' => [
          'id' => 'is_own',
          'plugin_id' => 'is_own',
          'settings' => [
            'query' => TRUE,
          ],
        ],
      ],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ])->save();

    $admin_user = $this->drupalCreateUser([
      'access administration pages',
      'access content overview',
      'administer nodes',
      'bypass node access',
      'view test_policy_with_and_rules content',
    ]);
    $admin_user->set('field_text', 'Matching text');
    $admin_user->save();

    $node_1 = $this->drupalCreateNode([
      'title' => 'Matching text (visible)',
      'type' => 'page',
      'status' => 1,
      'uid' => $admin_user->id(),
      'field_text' => 'Matching text',
    ]);

    $node_2 = $this->drupalCreateNode([
      'title' => 'Different owner (hidden)',
      'type' => 'page',
      'status' => 1,
      'uid' => 100,
      'field_text' => 'Matching text',
    ]);

    $node_3 = $this->drupalCreateNode([
      'title' => 'Different text and owner (hidden)',
      'type' => 'page',
      'status' => 1,
      'uid' => 100,
    ]);

    $this->assignAccessPolicy($node_1, 'test_policy_with_and_rules');
    $this->assignAccessPolicy($node_2, 'test_policy_with_and_rules');
    $this->assignAccessPolicy($node_3, 'test_policy_with_and_rules');

    $this->drupalLogin($admin_user);

    // Verify that only the content that the user can access is listed.
    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Matching text (visible)');
    $this->assertSession()->pageTextNotContains('Different owner (hidden)');
    $this->assertSession()->pageTextNotContains('Different text and owner (hidden)');
  }

  /**
   * Tests that a listing only displays entities that a user is allowed to view
   * when that policy has multiple rules using OR.
   */
  public function testContentListingWithMultipleORRules() {
    AccessPolicy::create([
      'id' => 'test_policy_with_or_rules',
      'label' => 'Test Access Policy with OR rules',
      'access_rules' =>  [
        'matching_text' => [
          'id' => 'matching_text',
          'plugin_id' => 'matching_text',
          'settings' => [
            'query' => TRUE,
          ],
        ],
        'is_own' => [
          'id' => 'is_own',
          'plugin_id' => 'is_own',
          'settings' => [
            'query' => TRUE,
          ],
        ],
      ],
      'access_rule_operator' => 'OR',
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ])->save();

    $admin_user = $this->drupalCreateUser([
      'access administration pages',
      'access content overview',
      'administer nodes',
      'bypass node access',
      'view test_policy_with_or_rules content',
    ]);
    $admin_user->set('field_text', 'Matching text');
    $admin_user->save();

    $node_1 = $this->drupalCreateNode([
      'title' => 'Matching text and owner (visible)',
      'type' => 'page',
      'status' => 1,
      'uid' => $admin_user->id(),
      'field_text' => 'Matching text',
    ]);

    $node_3 = $this->drupalCreateNode([
      'title' => 'Different text (visible)',
      'type' => 'page',
      'status' => 1,
      'uid' => $admin_user->id(),
      'field_text' => 'Different text',
    ]);

    $node_2 = $this->drupalCreateNode([
      'title' => 'Different owner (visible)',
      'type' => 'page',
      'status' => 1,
      'uid' => 100,
      'field_text' => 'Matching text',
    ]);

    $node_4 = $this->drupalCreateNode([
      'title' => 'Different text and owner (hidden)',
      'type' => 'page',
      'status' => 1,
      'uid' => 100,
      'field_text' => 'Different text',
    ]);

    $this->assignAccessPolicy($node_1, 'test_policy_with_or_rules');
    $this->assignAccessPolicy($node_2, 'test_policy_with_or_rules');
    $this->assignAccessPolicy($node_3, 'test_policy_with_or_rules');
    $this->assignAccessPolicy($node_4, 'test_policy_with_or_rules');

    $this->drupalLogin($admin_user);

    // Verify that only the content that the user can access is listed.
    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('Matching text and owner (visible)');
    $this->assertSession()->pageTextContains('Different text (visible)');
    $this->assertSession()->pageTextContains('Different owner (visible)');
    $this->assertSession()->pageTextNotContains('Different text and owner (hidden)');
  }

  /**
   * Tests that a listing displays entities with Access Rules a user can bypass.
   */
  public function testContentListingWithBypassAccessRulesEnabled() {
    AccessPolicy::create([
      'id' => 'matching_text',
      'label' => 'Matching Text Access Policy',
      'access_rules' =>  [
        'matching_text' => [
          'id' => 'matching_text',
          'plugin_id' => 'matching_text',
          'settings' => [
            'query' => TRUE,
          ],
        ],
      ],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ])->save();

    $admin_user = $this->drupalCreateUser([
      'access administration pages',
      'access content overview',
      'administer nodes',
      'bypass node access',
      'view matching_text content',
      'bypass matching_text access rules',
    ]);
    $admin_user->set('field_text', 'Matching text');
    $admin_user->save();

    $node_1 = $this->drupalCreateNode([
      'title' => 'First',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Matching text',
    ]);

    $node_2 = $this->drupalCreateNode([
      'title' => 'Second',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Restricted',
    ]);

    $this->assignAccessPolicy($node_1, 'matching_text');
    $this->assignAccessPolicy($node_2, 'matching_text');

    $this->drupalLogin($admin_user);

    // Verify that content for which a user can bypass Access Rules is listed.
    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('First');
    $this->assertSession()->pageTextContains('Second');
  }

  /**
   * Tests that an access rule without a query doesn't accidentally break the
   * query.
   */
  public function testContentListingWithAccessRulesQueryMissing() {
    AccessPolicy::create([
      'id' => 'rules_with_no_query',
      'label' => 'Test Access Policy with no query rules',
      'access_rules' =>  [
        'access_rule_text' => [
          'id' => 'access_rule_test',
          'plugin_id' => 'access_rule_test',
          'settings' => [
            'query' => TRUE,
          ],
        ],
      ],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ])->save();

    $admin_user = $this->drupalCreateUser([
      'access administration pages',
      'access content overview',
      'administer nodes',
      'bypass node access',
      'view rules_with_no_query content',
    ]);
    $admin_user->set('field_text', 'Does not match');
    $admin_user->save();

    $node_1 = $this->drupalCreateNode([
      'title' => 'First',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'First',
    ]);

    $node_2 = $this->drupalCreateNode([
      'title' => 'Second',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Second',
    ]);

    $this->assignAccessPolicy($node_1, 'rules_with_no_query');
    $this->assignAccessPolicy($node_2, 'rules_with_no_query');

    $this->drupalLogin($admin_user);

    // If query is enabled for the rules then these would not be visible.
    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('First');
    $this->assertSession()->pageTextContains('Second');
  }

  /**
   * Tests that a listing displays entities when Access Rule query disabled.
   */
  public function testContentListingWithAccessRulesQueryDisabled() {
    AccessPolicy::create([
      'id' => 'rules_with_disabled_query',
      'label' => 'Test Access Policy with disabled rules',
      'access_rules' =>  [
        'matching_text' => [
          'id' => 'matching_text',
          'plugin_id' => 'matching_text',
          'settings' => [
            'query' => FALSE,
          ],
        ],
      ],
      'target_entity_type_id' => 'node',
      'target_bundles' => ['page'],
    ])->save();

    $admin_user = $this->drupalCreateUser([
      'access administration pages',
      'access content overview',
      'administer nodes',
      'bypass node access',
      'view rules_with_disabled_query content',
    ]);
    $admin_user->set('field_text', 'Does not match');
    $admin_user->save();

    $node_1 = $this->drupalCreateNode([
      'title' => 'First',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'First',
    ]);

    $node_2 = $this->drupalCreateNode([
      'title' => 'Second',
      'type' => 'page',
      'status' => 1,
      'field_text' => 'Second',
    ]);

    $this->assignAccessPolicy($node_1, 'rules_with_disabled_query');
    $this->assignAccessPolicy($node_2, 'rules_with_disabled_query');

    $this->drupalLogin($admin_user);

    // If query is enabled for the rules then these would not be visible.
    $this->drupalGet('admin/content');
    $this->assertSession()->statusCodeEquals(200);
    $this->assertSession()->pageTextContains('First');
    $this->assertSession()->pageTextContains('Second');
  }

}
